package IS.hi.logir.cryptonite;

import java.math.BigInteger;
import java.io.Serializable;

/**
 * The Caesar algorithm is supposedly the one Julius Caesar used by
 * hand many centuries ago. As you can imagine, this is <b>NOT A STRONG
 * CIPHER</b>, but included only to show how to write a very simple
 * Key class for the Cryptonite package.
 * <p>
 * Data is encrypted byte-for-byte by adding a constant value to it and
 * taking the 8 lowest order bits.
 * 
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public final class CaesarKey extends Key implements Serializable, Fingerprintable{

    private byte shift;
    
    /**
     * Create a new Caesar key with the specified shift. It will be marked
     * as belonging to ownerName with e-mail address ownerMail. */
    public CaesarKey(byte shift, String ownerName, String ownerMail) {
        super(ownerName,ownerMail);
        this.shift = shift;
    }

    /**
     * Create a pair of public/private keys. They will be marked
     * as belonging to ownerName with e-mail address ownerMail. */
    public static KeyPair createKeys(String ownerName, String ownerMail){
        if (random==null)
            random=new RandomSpinner();
        byte shift = (byte)(random.nextInt()%255+1);
        
        Key k=new CaesarKey(shift, ownerName, ownerMail);
        return new KeyPair(k,k);  // The same key is used for encryption and decryption.
    }

    /**
     * Create a pair of public/private keys. They will be marked
     * as belonging to ownerName with e-mail address ownerMail. */    
    public static KeyPair createKeys(int notUsed, String ownerName, String ownerMail){
        return createKeys(ownerName, ownerMail);
    }

    /** Return the size of the key in bits. */
    public int getSize(){
        return 8;
    }

    /**
     * Return the key's MD5 fingerprint.
     *
     * @see IS.hi.logir.cryptonite.MD5State
     * @see IS.hi.logir.cryptonite.Fingerprintable
     */
    protected final Fingerprint calcFingerprint(){
        byte[] bN=getOwnerName().getBytes();
        byte[] bM=getOwnerMail().getBytes();
        byte[] b = new byte[4+bN.length+bM.length];
        b[0] = shift;
        System.arraycopy(bN,0, b,1          , bN.length);
        System.arraycopy(bM,0, b,1+bN.length, bM.length);
        return MD5State.calcFingerprint(b);
    }

    /** Return the name of the algorithm used by this key. */
    public String algorithmName(){
        return ("Caesar");
    }

    /** Return a string representation of this key. */
    public String toString(){
	// You should of course never write the actual key to the
	// screen here, since it may be seen by a passer by (or grabbed
	// from the screen). However, CaesarKey is so bad that it doesn't
	// matter anyway.
	return "Caesar key ("+shift+     
            ")\n    from "+getOwnerName()+" ("+getOwnerMail()+")";
    }

    /** Return true if the two keys are eqivalent. */
    public final boolean equals(Object o){
        if (o.getClass() != this.getClass())
            return false;
        return shift==((CaesarKey)o).shift;
    }

    /** 
     * Returns true if this and key are a matched pair
     * of public/private keys. */
    public final boolean matches(Key key){
        return equals((Key)key);
    }
    
    /**
     * Encrypt the first sourceLength bytes of source with
     * this key and return an array with the encrypted data. */
    public final void encrypt(Data source, Data dest){
	dest.enlarge(source.n);
	for (int i=0; i<source.n; i++)
            dest.buf[i] = (byte)(source.buf[i]+shift);
	dest.n=source.n;
    }
     
    /**
     * Decrypt the data from source with this key and put the plain data
     * in dest. The source.buf and dest.buf arrays must be different. */
     public void decrypt(Data source, Data dest){
	 dest.enlarge(source.n);
	 for (int i=0; i<source.n; i++)
	     dest.buf[i] = (byte)(source.buf[i]-shift);
	 dest.n=source.n;
     }

}
