package IS.hi.logir.cryptonite;

import java.io.Serializable;

/**
 * The Data class is hods an ordered pair (n,buf), where buf is an array of
 * bytes containing data in the first n bytes. This is used to return data
 * from methods where we want to re-use arrays if possible and only
 * re-allocate them if we need to them.
 * <p>
 * For example, the call <tt>Key.encrypt(buf, dest)</tt> takes a parameter
 * of type Data in dest and if the encrypted array is no larger than dest.buf,
 * it just writes the encrypted data to it. Otherwise dest.buf is
 * enlarged. In either case, dest.n is the actual number of bytes of
 * encrypted data.
 * 
 * @see IS.hi.logir.cryptonite.Key
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class Data implements Serializable, Fingerprintable {
    // This object holds n bytes of data in buf[0..n-1].
    public int n;
    public byte[] buf;
    
    // All instances created with Data() use this empty array rather than
    // creating an empty array each.
    static private byte[] EMPTY = new byte[0];
    
    /** Create a new empty Data object. */
    public Data(){
	n=0;
	buf=EMPTY;
    }
    
    /**
     * Create a new empty data object that can hold
     * m bytes without being enlarged. */
    public Data(int m){
	n=0;
	buf=new byte[m];
    }
    
    /**
     * Create a new data object around the bytes in buffer.
     * The buffer must contain at least m bytes. */
    public Data(byte[] buffer){
	buf=buffer;
	n=buf.length;
    }
    
    /**
     * Create a new data object around the first m bytes
     * in buffer. The buffer must contain at least m bytes. */
    public Data(byte[] buffer, int m){
	buf=buffer;
	n=m;
    }
    
    /**
     * Make sure that the object can hold at least m bytes.
     * This clears the object. */
    public synchronized void enlarge(int m){
	n=0;
	if (buf.length<m)
	    buf=new byte[m];
    }
    
    /**
     * Make sure that the object can hold at least m bytes,
     * without destroying the data it contains. */
    public synchronized void enlargeSafe(int m){
	if(buf.length<m){
	    byte[] old=buf;
	    buf=new byte[m];
	    System.arraycopy(old,0, buf,0, n);
	}
    }

    /**
     * Reallocate the buffer and make it m bytest large.
     * This clears the object. */
    public synchronized void reallocate(int m){
	n=m;
	buf=new byte[n];
    }

    /**
     * Resize the buffer so it can hold as close to m bytes 
     * as possible without losing data. */
    public synchronized void resizeSafe(int m){
	if (m<n)
	    m=n;
	if (m==buf.length)
	    return;
	byte[] old = buf;
	buf = new byte[m];
	System.arraycopy(old,0, buf,0, n);
    }
    
    public Fingerprint getFingerprint(){
	MD5State md5 = new MD5State();
	md5.update(buf,0,n);
	return md5.calculate();
    }
    
    /* IMPLEMENT TO WRITE AS LITTLE DATA AS POSSIBLE TO A DATA OUTPUT STREAM 
    private void writeObject(java.io.ObjectOutputStream stream)
	throws IOException;
    private void readObject(java.io.ObjectInputStream stream)
	throws IOException, ClassNotFoundException;
    */
}
