package IS.hi.logir.cryptonite;

import java.math.BigInteger;

import java.io.InputStream;
import java.io.FilterInputStream;
import java.io.ObjectInputStream;
import java.io.ByteArrayInputStream;

import java.io.IOException;
import java.io.InterruptedIOException;

/**
 * This class is a FilterInputStream that decrypts a stream of bytes similar to that
 * generated by EncryptStream. The format of the encrypted stream is shown in the
 * documantation for EncryptStream.
 *
 * @see IS.hi.logir.cryptonite.Key
 * @see IS.hi.logir.cryptonite.EncryptStream
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
 
public final class DecryptStream extends FilterInputStream {
    // An instance of DecryptStream can be in one of 5 states:
    // 1) blockType == UNKNOWN
    //    We haven't read the first integer of the next block and
    //    therefore don't know what kind of block it is. This also
    //    means that data.buf[dataPos..data.n-1] contains the
    //    decrypted but unread bytes in the stream.
    // 2) blockType == SESSION_KEY
    //    We're reading a session key block.
    // 3) blockType == KEY
    //    We're reading a key identifier block.
    // 4) blockType == DATA
    //    We're reading a block of ciphertext.
    // 5) blockType == EOF
    //    The end of the underlying stream has been reached, and there is
    //    no decrypted data left for the user.
    //
    // If blockType != UNKNOWN then blockSize is the size of the block being
    // read in bytes, source.buf[0..source.n-1] is the part of the block
    // already read and source is large enough to contain the rest.

    private Data data=new Data();
    private Data source=new Data();
    private int dataPos;      
 
    private int blockSize;
    private int blockType;
 
    private final int UNKNOWN     = -3;   // We haven't seen what kind of bock follows
    private final int EOF         = -2;   // We've reached the endo of the stream
    private final int SESSION_KEY = -1;   // A block containing a session key
    private final int KEY         =  0;   // A block containing a key identifier
    private final int DATA        =  1;   // A block of ciphertext

    private Key key;         // The private used for decryption.
    private Key sKey;        // The current session key.
    private KeyRing keyRing; // The key-ring used to find the keys that the encrypted stream claims to use if they are different from the key field.

    /**
     * Create a new stream which decrypts data from in. If a key is specified, it will be
     * used to decrypt the data. If the encrypted stream contains a key identifier for another
     * key, then keyRing is used to find the key.
     * <p>
     * If key is null, then the encrypted data is assumed to contain a key identifier
     * to specify a key in keyRing.
     * <p>
     * If keyRing is null, then the encrypted data is assumed to be incrypted with the
     * specified key.
     * <p>
     * Setting both key and keyRing to null is an error.
     */
    public DecryptStream(InputStream in, Key key, KeyRing keyRing){
        super(in);
        this.keyRing=keyRing;
        this.key=key;
	sKey=key;

	dataPos = 0;
        blockType = UNKNOWN;
    }

    /** Closes this input stream and releases any system resources associated with the stream. */
    public synchronized void close() throws IOException{
        super.close();
    }

    /** 
     * Returns the number of bytes that can be read from
     * this input stream without blocking. */
    public synchronized int available() throws IOException {
        if (blockType==UNKNOWN && data.n!=dataPos)
            // We're not currently reading a block from the underlying stream
            return data.n-dataPos;
        getBlock(false);
        if (blockType==UNKNOWN)
            // We couldn't start reading the next block
            return 0;
        else
            // We've just started reading a new block.
            return data.n-dataPos;     
    }

    /**
     * Reads the next block of encrypted data from the input stream.
     * If waitIfBlocked is true the method will block if data can
     * not be read from the underlying stream immediately. */
     private synchronized void getBlock(boolean waitIfBlocked) throws IOException, BadKeyClassException {
	 if (blockType==EOF)
	     // We've reached the end of the stream
	     return;
         data.n=dataPos=0;
     
         if(blockType==UNKNOWN){
	     // We haven't read the next block-size
	     if(!waitIfBlocked && (in.available()<4))
		 return;
             int m = (in.read()<<24) | (in.read()<<16) | (in.read()<<8);
             int n=in.read();
             if(n==-1) {
		 // We've reached the end of the stream.
		 blockType=EOF;
                 return;
             }
             m = m | n;
             // m is the first int from the next block.
             if (m>0){
		 // We have a cipher block of m bytes.
		 blockType=DATA;
		 blockSize=m;
	     } else if (m==0)
		 // We have a fixed-length plain key ID block
	         blockType=KEY;
             else if (m<0){
                 // We have an encrypted session key of -m bytes
                 blockType=SESSION_KEY;
		 blockSize=-m;
	     }
             source.n=0;
         }

         if(blockType==KEY){
	     if (source.n==0)
		 source.enlarge(16);
	     int bytesToRead = waitIfBlocked ? 16-source.n : Math.min(16-source.n, in.available());
	     int bytesRead = in.read(source.buf, source.n, bytesToRead);
	     if (bytesRead<0)
		 throw new InterruptedIOException("A key-identifier block was prematurely terminated.");
	     source.n+=bytesRead;
	     if (source.n==16){
		 boolean noSKey = (key==null) || (key.equals(sKey));
		 // We've finished reading the key identifier block.
		 Fingerprint fingerprint = new Fingerprint("MD5", source.buf, source.n);
                 if ((key==null) || (!fingerprint.equals(key.getFingerprint()))){
		    // It's not our old key.
                    if (keyRing==null)
                        throw new DecryptException("An encrypted data-stream needs another key, but no KeyRing was specified.");
                    key = keyRing.getPri(fingerprint);
                    if (key==null)
                        throw new KeyException("An encrypted data-stream needs a key which was not found in the KeyRing.");
                    if(noSKey)
                        sKey=key;
                }
                blockType=UNKNOWN;
                source.n=0;
            }   
        } 

        if(blockType==SESSION_KEY || blockType==DATA){
            // We need to read an encrypted block.
	    if (source.n==0)
                // we've got to make sure the buffer is large
		// enough to hold the encrypted data.
		source.enlarge(blockSize);
            
            if (waitIfBlocked){
                // keep trying until we've read the whole lot
                while(source.n<blockSize){
		    // we've read source.n bytes of cipher text.
                    int bytesRead = in.read(source.buf,source.n,blockSize-source.n);
                    if(bytesRead==-1)
                        throw new InterruptedIOException("Encrypted data-stream was prematurely terminated.");
                    source.n += bytesRead;
                }
            } else {
                // read as much as we can without blocking
                int bytesRead = in.read(source.buf,source.n, Math.min(blockSize-source.n, in.available()));
                if(bytesRead==-1)
                    throw new InterruptedIOException("Encrypted data-stream was prematurely terminated.");
		source.n += bytesRead;
            }
	}
     
        if(blockType==SESSION_KEY){		
            // We need to read an encrypted session key.
            
            if(source.n==blockSize){
		// we've completed reading the block		    

                key.decrypt(source, data);
                // data holds a plain, serialized session key.
     
                ByteArrayInputStream byteIn = new ByteArrayInputStream(data.buf);
                ObjectInputStream objIn = new ObjectInputStream(byteIn);
      
                if (sKey==null)
                    throw new KeyException("No key has been specified to decrypt session-keys from this DecryptStream.");
                try{
                    sKey = (Key)objIn.readObject();
		} catch (Exception e){
		    System.err.println(e);
		    e.printStackTrace();
		    throw new BadKeyClassException("The class of the session key is unrecognized.");
		}
                // sKey is the new session key.
                blockType=UNKNOWN;
                source.n=0;
                data.n=0;
            }
        }
     
        if(blockType==DATA){
            if(source.n==blockSize){
                // we've completed reading the block
                if (sKey==null)
                    throw new KeyException("No key has been specified to decrypt data in this DecryptStream.");
                sKey.decrypt(source, data);
                blockType=UNKNOWN;
                source.n=0;
            }
        }
    }

    /** Reads the next byte of data from this input stream. */
    public synchronized int read() throws IOException{
        if(blockType!=UNKNOWN || (dataPos==data.n))
            getBlock(true);
        if (blockType==EOF)
            return -1;
        return (int)data.buf[dataPos++] & 0xff;
    }

    /** Reads up to b.length bytes of data from this input stream into an array of bytes. */
    public synchronized int read(byte[] b) throws IOException{
        return read(b,0,b.length);
    }

    /** Reads up to len bytes of data from this input stream into an array of bytes. */
    public synchronized int read(byte[] b, int off, int len) throws IOException, CryptoException {
        if(blockType!=UNKNOWN)
            getBlock(true);
        int left=len;
        while(left>0){
            // we still need to copy len bytes of decrypted data
            if(dataPos==data.n)
		// nothing in buffer
                getBlock(true);
            if(blockType==EOF)
		// uh-oh, end of stream reached.
                return (len==left) ? -1 : len-left;
            int bytesToCopy=Math.min(left,data.n-dataPos);
            System.arraycopy(data.buf,dataPos, b,off, bytesToCopy);
            dataPos+=bytesToCopy;;
            off+=bytesToCopy;
            left-=bytesToCopy;
        }
        return len-left;
    }

    /** Returns false, signifying that mark and reset is not supported. */
    public synchronized boolean markSupported(){ return false; }

    /** Marks the current position in this input stream: <b>NOT SUPPORTED</b>. */
    public synchronized void mark(int readLimit){ }

    /**
     * Repositions this stream to the position at the time the mark
     * method was last called on this input stream.: <b>NOT SUPPRTED</b> */
    public synchronized void reset(){ }
}
