package IS.hi.logir.cryptonite;

import java.math.BigInteger;

import java.io.OutputStream;
import java.io.FilterOutputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;  // temporary...
import java.io.ObjectInputStream;  // temporary...

import java.io.IOException;

/**
 * This class is a FilterOutputStream that encrypts everything passing through it.
 * The format of the encrypted stream is as follows:<blockquote>
 *   <b>encrypted stream</b>      = &lt;key identifier&gt;[block [... [block] ...]]<br>
 *   <b>block</b>                 = &lt;key identifier | session key | encrypted block&gt;<br>
 *   <b>key identifier</b>        = &lt;0x00000000&gt;&lt;MD5 fingerprint&gt;<br>
 *   <b>session key</b>           = &lt;negative length of block&gt;&lt;serialized and encrypted Key object&gt;<br>
 *   <b>key-class name length</b> = 32-bit integer<br>
 *   <b>key-class name</b>        = string<br>
 *   <b>cipher block</b>          = &lt;decrypted size&gt;&lt;cipher chunk&gt;[cipher chunk[... [cipher chunk]... ]]
 *   <p>
 *   The definition of <b>key</b> depends on the particular key class.
 *   <p>
 *   The definition of <b>cipher chunk</b> and a more detailed explanation of
 *   <b>cipher block</b> is shown in the documantation of the class Key.
 * </blockquote>
 * The last <b>key identifier</b> before a <b>session key</b> block
 * identifies the key used to encrypt the <b>session key</b>.
 * <p>
 * The last <b>session key</b> before a <b>cipher block</b> contains the
 * key used to decrypt that block. If no <b>session key</b> block has
 * appeared, then the key from the last <b>key identifier</b> block should
 * be used in stead.
 *
 * @see IS.hi.logir.cryptonite.Key
 * @see IS.hi.logir.cryptonite.DecryptStream
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
 
public final class EncryptStream extends FilterOutputStream {
    
    private Key key;     // The public key used for encryption
    private Key sKey;    // The session-key currently in use (=?= key)
    
    private Data buf;    // Data to be encrypted is in buf[0..n-1]
    private int bufSize; // Bytes to encrypt at once (unless flushed)
  
    // The buffer size we want is stored in idealBufSize. This will be
    // rounded up to the next multiple of key.blockSize() to save space
    // and used os the size of the output buffer.
    private int idealBufSize; 
    
   /**
    * Create a new stream which encrypts data written to it and writes it to
    * out. The specified key is used to encrypt the data in blocks of up to
    * most idealBufSize bytes.
    */
    public EncryptStream(OutputStream out, Key key, int idealBufSize) throws IOException {
	super(out);
	init(bufSize, key);
    }
 
   /**
    * Create a new stream which encrypts data written to it and writes it to
    * out. The specified key is used to encrypt the data in blocks of up to
    * 2KB. */
    public EncryptStream(OutputStream out, Key key) throws IOException {
        super(out);
	init(1024, key);
    }

    private void init(int bufSize, Key key) throws IOException{
	this.key=key;
	sKey=key;
	idealBufSize=bufSize;	
	buf=new Data();
	resizeBuffer();
	writeKey();
    }
 
   /**
    * Closes this output stream and releases any system
    * resources associated with the stream. */
    public synchronized void close() throws IOException{
        flush();
        out.close();
    }
 
    // encrypt and write the content of buf
    private synchronized void writeBuffer() throws IOException{
        if (buf.n>0){
	    Data cipher = new Data();
            sKey.encrypt(buf,cipher);
	    out.write(cipher.n >> 24);
	    out.write(cipher.n >> 16);
	    out.write(cipher.n >>  8);
	    out.write(cipher.n      );
            out.write(cipher.buf, 0, cipher.n);	    
            buf.n=0;
        }
    }
    
    /**
     * Flushes this output stream and forces any buffered
     * output bytes to be encrypted and written. */
    public synchronized void flush() throws IOException{
        writeBuffer();
        out.flush();
    }

    /** Encrypts and writes b.length bytes to the stream. */
    public synchronized void write(byte[] b) throws IOException{
        write(b,0,b.length);
    }

    /**
     * Encrypts and writes len bytes from the specified byte array starting at
     * offset off to this output stream. */
    public synchronized void write(byte[] b, int off, int len) throws IOException{
        while(len>0){
            // We have written all the data except b[off..off+len-1]
            int chunkLen=Math.min(len,bufSize-buf.n);
            System.arraycopy(b,off, buf.buf,buf.n, chunkLen);
            off+=chunkLen;
            len-=chunkLen;
            buf.n+=chunkLen;
            if(buf.n==bufSize)
                writeBuffer();
        }
    }

    /** Encrypt and write the specified byte to the stream. */
    public synchronized void write(int b) throws IOException{
        buf.buf[buf.n++]=(byte)b;
        if(buf.n==bufSize)
            writeBuffer();
    }

    private void resizeBuffer() throws IOException {
	writeBuffer();
	    
	int blockSize = sKey.blockSize();
	if (idealBufSize % blockSize == 0)
	    bufSize=idealBufSize;
	else
	    bufSize=blockSize*(1+idealBufSize/blockSize);
	// this is OK, since we've just written the buffer above.
	buf.enlarge(bufSize);
    }
 
    /**
     * Sets the key to use for encrypting data. If a key identifier block has
     * previously been written, another one will be written to reflect the
     * change. */
    public synchronized void setKey(Key key) throws IOException{
        if (key.equals(this.key))
            return;
        boolean noSKey = this.key.equals(sKey);

	writeBuffer();
        this.key=key;
        if (noSKey){
	    sKey=key;	
            resizeBuffer();
	}
     }
        
     // Write a key identifier block. 
     private synchronized void writeKey() throws IOException{
	 writeBuffer();
	 out.write(0);
	 out.write(0);
	 out.write(0);
	 out.write(0);
	 out.write(key.getFingerprint().getBytes());
     }
     
     /**
      * Set the session key to use for this stream. Session keys can be
      * use symmetric ciphers, which tend to be much faster than the
      * public key ciphers used in the constructor. The session key will
      * be encrypted with the public key and written to the stream.
      */
     public synchronized void setSessionKey(Key sk) throws IOException{
         if (!sk.matches(sk))
	     throw new KeyException("Symmetric key expected");
	     
	 writeBuffer();

	 ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
	 ObjectOutputStream objOut = new ObjectOutputStream(byteOut);
	 objOut.writeObject(sk);
	 objOut.flush();
	 
	 byte[] buffer = byteOut.toByteArray();
	 byte[] cipher = key.encrypt(buffer);
	 // cipher now contains the encrypted and serialized session key.

	 int blockSize = -cipher.length;
	 out.write(blockSize >> 24);
	 out.write(blockSize >> 16);
	 out.write(blockSize >>  8);
	 out.write(blockSize      );
	 out.write(cipher);

	 sKey=sk;
     }
}
