package IS.hi.logir.cryptonite;

import java.io.Serializable;

/**
 * This class is used to hold a fingerprint of a particular data buffer.
 * The idea is to calculate a fingerprint in such a way that it is extremely
 * difficult to create another buffer that gives the same fingerprint. If
 * that buffer also has to match some other criteria, such as being a valid 
 * text file in a particular language, then it becomes next to impossible.
 * <p>
 * All this depends on the hash function used to create the fingerprint
 * being a good one. Fingerprints are created by the various subclasses
 * of FingerprintState, so you should look there for information about
 * a particular hash function.
 * <p>
 * You should be very careful about saving fingerprints. If a fingerprint
 * is not calculated at run-time, then it should not be trusted, since it
 * may have been tampered with in storage or transit. This is why the
 * private fingerprint field of the Key class is marked as transient and
 * re-calcualted when the class is deserialized.
 * <p>
 *
 * @see IS.hi.logir.cryptonite.FingerprintState
 * @see IS.hi.logir.cryptonite.Signature
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class Fingerprint implements Serializable{

    /**
     * Creates a fingerprint for the buffer buf using the named
     * fingerprint-class.
     * <p>
     * A fingerprint class with the specified name must previously have
     * been registered by calling the registerIncludedClasses method
     * on the FingerprintState class. This is automatically done for
     * classes included with cryptonite but any other classes must be
     * registered manually.
     */
    public static Fingerprint get(String name, byte[] buf) throws BadFingerprintClassException {
        FingerprintState s = FingerprintState.createState(name);
        s.update(buf);
        return s.calculate();
    }


    // =================================================================== //
    
    /** fp holds the actual bytes of the fingerprint value. */
    protected byte[]fp=null;

    /**
     * algorithm holds the name of the algorithm
     * used to create this fingerprint. */
    protected String algorithm=null;

    /** 
     * Creates a new Fingerprint object given algorithm name and
     * the hash value from fp[0..n-1]. */
    public Fingerprint(String algorithm, byte[] fp, int n){
        this.algorithm = algorithm;
        this.fp = new byte[n];
        System.arraycopy(fp,0, this.fp,0, n);
    }
    
    /** 
     * Creates a new Fingerprint object given algorithm name and
     * the hash value from fp. */
    public Fingerprint(String algorithm, byte[] fp){
        this.algorithm = algorithm;
        this.fp = new byte[fp.length];
        System.arraycopy(fp,0, this.fp,0, fp.length);
    }
    
    /** Return the name of the algorithm used for this hash. */
    public String getName(){
        return algorithm;
    }

    /** Return an array of the bytes in the fingerprint. */
    public byte[] getBytes(){
        byte[] fp=new byte[this.fp.length];
        System.arraycopy(this.fp,0, fp,0, this.fp.length);
        return fp;
    }

    /** return true if obj is a fingerprint equal to this. */
    public boolean equals(Object obj){
        if (obj.getClass()!=this.getClass())
            return false;
        // It's the same class
        
        Fingerprint f=(Fingerprint)obj;
        if (!f.algorithm.equals(algorithm))
            return false;
        // It's the same algorithm.
        
        for (int i=fp.length-1; i>=0; i--)
            // this.fp[j] == fp[j] for j=0, 1, ..., i-1.
            if (fp[i]!=f.fp[i])
                return false;

        return true;
    }

    public int hashCode(){
        int h=0;
        for (int i=0; i<fp.length; i++)
            h = h ^ ( fp[i] << (i%4)*8 );
        return h;
    }

    private static char[] nibble = { '0', '1', '2', '3', '4', '5', '6', '7',
                                     '8', '9', 'a', 'b', 'c', 'd', 'e', 'f',
                                     '0', '1', '2', '3', '4', '5', '6', '7',
                                     '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};

    /** Return a string represesntation of this fingerprint. */
    public String toString(){
        if (fp==null)
            return "<unassigned>";
        StringBuffer sb = new StringBuffer(32);
        for (int i=0; i<fp.length; i++){
            sb.append(nibble[(fp[i]>>4)&15]);
            sb.append(nibble[fp[i]&15]);
        }
        return sb.toString();
    }
}
