package IS.hi.logir.cryptonite;

import java.io.Serializable;
import java.util.Hashtable;
import java.lang.reflect.Method;

/**
 * An object of this class holds the status of a fingerprint still being
 * calculated.
 * <p>
 * A fingerprint state object can be repeatedly updated with data. At
 * any time a Fingerprint object can be requested for the data that has
 * then been added to the fingerprint state.
 * <p>
 * Any subclass of FingerprintState should override the abstract methods
 * in FingerprintState and add a method:
 * <blockquote><tt>
 *   public static FingerprintState createState() {...}
 * </tt></blockquote>
 * which returns a new instance of the subclass.
 * 
 * @see IS.hi.logir.cryptonite.MD5Fingerprint
 * @see IS.hi.logir.cryptonite.FingerprintStatus
 * @see IS.hi.logir.cryptonite.Signature
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public abstract class FingerprintState implements Serializable {

    // fingerprintClasses.find(S) is the Class object for the class
    // used to implement fingerprints for the algorithms called S.
    private static Hashtable fingerprintClasses = new Hashtable();
    
    /**
     * Add the class c to the list of available fingerprint classes. It
     * is registered with the specified name. Each non-abstract sub-class
     * of Fingerprint should call Fingerprint.addHashClass to make its
     * existance known. */
    public static void addFingerprintClass(String name, Class c){
	fingerprintClasses.put(name,c);
    }
    
    static boolean classesRegistered;  // True if the included fingerprint classes have been registered.

    private static void registerIncludedClasses() throws BadFingerprintClassException{
	if (classesRegistered)
	    return;
	try{
	    addFingerprintClass("MD5", Class.forName("IS.hi.logir.cryptonite.MD5State"));
	} catch (ClassNotFoundException e) {
	    throw new BadFingerprintClassException("Unable to get Class object for MD5State");
	}
	classesRegistered=true;	
    }

    /**
     * Create a FingerprintState object for the named fingerprint
     * algorithm.
     */
    public static FingerprintState createState(String name) throws BadFingerprintClassException {
	registerIncludedClasses();
	
	Class c=(Class)(fingerprintClasses.get(name));
	if (c==null)
	    throw new BadFingerprintClassException("There is no fingerprint class registered with the name "+name);
	// c is the Class object for the named fingerprint class
	FingerprintState s;
	try {
	    Class[] paramClass = { };
	    Method m=c.getMethod("createState", paramClass );
	    Object[] param = { };
	    s = (FingerprintState)m.invoke(null, param);
	} catch (Exception e) {
	    throw new BadFingerprintClassException("Unable to create a FingerprintState object for "+name);
        }
	// s is a new FingerprintState object of the appropriate class..
	return s;	
    }    
    
    /** Reset the state. */
    public abstract void reset();
     
    /**
     * Update the fingerprint state with the bytes from
     * buf[offset, offset+length-1]. */
    public abstract void update(byte[] buf, int offset, int length);
    
    /** Update the fingerprint state with the bytes from buf. */
    public void update(byte[] buf){
	update(buf, 0, buf.length);
    }
    
    /** Update the fingerprint state with the characters from s. */
    public void update(String s){
	update(s.getBytes());
    }
    
    /** 
     * Return a fingerprint for the curret state, without
     * destroying the state. */
    public abstract Fingerprint calculate();
    
}
