package IS.hi.logir.cryptonite;

import java.math.BigInteger;
import java.io.Serializable;
import java.util.Random;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.lang.reflect.Method;

/**
 * This class is the ancestor of classes to hold one public or private key.
 * It's descendants also handle the actual encryption and decryption of data.
 * <p>
 * Subclasses of Key should be "read-only" in the same sense as the String
 * class and not implement any methods to change the state of the object.
 * <p>
 * The format of the ciphertext is as follows:<blockquote>
 *   <b>cipher block</b>=&lt;decrypted size&gt;&lt;cipher chunk&gt;[cipher chunk[... [cipher chunk]... ]]
 *   <p>
 *   Each cipher chunk is a whole number of bytes. The excact number of bytes
 *   depends on the key used to encrypt the data.
 *   <p>
 *   There may not be enough plain data to fill the last chunk, in which case
 *   the last chunk is calculated as if 0's were added to the end of the
 *   plaintext.
 * </blockquote>
 * <p>
 * 
 * @see IS.hi.logir.cryptonite.KeyPair
 * @see IS.hi.logir.cryptonite.KeyRing
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public abstract class Key implements Serializable, Fingerprintable {

    // KEY CLASS LIBRARIAN
    
    private static Hashtable keyClasses=new Hashtable();

    /**
     * Add the class c to the list of available key classes. This allows
     * the creation of keys of this class by name.
     * <p>
     * All regstered classes must be sub-classes of the Key class and
     * implement a <tt>static KeyPair createKeys(int size, String
     * ownerName, String ownerMail)</tt> method similar to that in the
     * RsaKey class.
     *
     * @see IS.hi.logir.cryptonite.Key#create
     * @see IS.hi.logir.cryptonite.RsaKey#createKeys
     */
    public static void addKeyClass(String name, Class c){
        keyClasses.put(name,c);
    }

    private static boolean isRegistered=false;

    /**
     * Register all the included classes
     * so they can be created by name. */
    private static void registerIncludedClasses() throws BadKeyClassException {
        if (isRegistered)
            return;
        try{
            addKeyClass("RSA", Class.forName("IS.hi.logir.cryptonite.RsaKey"));
        } catch (ClassNotFoundException e) {
            throw new BadKeyClassException("Unable to get Class object for RsaKey");
        }
        try{
            addKeyClass("Tea", Class.forName("IS.hi.logir.cryptonite.TeaKey"));
        } catch (ClassNotFoundException e) {
            throw new BadKeyClassException("Unable to get Class object for TeaKey");
        }
        try{
            addKeyClass("Caesar", Class.forName("IS.hi.logir.cryptonite.CaesarKey"));
        } catch (ClassNotFoundException e) {
            throw new BadKeyClassException("Unable to get Class object for CaesarKey");
        }
        isRegistered=true;
    }
    
    /**
     * Create a new KeyPair for the named key class with the specified size.
     * <p>
     * A key class with the specified name must previously have been
     * registered by calling addKeyClass. This is automatically done
     * for those classes included with cryptonite but any other classes
     * must be registered manually.
     * <p>
     * The meaning of the size parameter depends on the key class. If more
     * than one parameter is needed to fully specify the kind of key to
     * create, a default will be used. If this is not acceptable, use the
     * constructor for the particular key class.
     * <p>
     * The keys will be marked as belonging to ownerName with e-mail
     * address ownerMail. 
     *
     * @see IS.hi.logir.cryptonite.Key#addKeyClass
     */
    public static KeyPair create(String name, int size, String ownerName, String ownerMail) throws BadKeyClassException{
        registerIncludedClasses();
        
        Class c=(Class)keyClasses.get(name);
        if (c==null)
            throw new BadKeyClassException("There is no encrytion algorithm called "+name);
        // c is the Class object for the named key class
        KeyPair kp;
        try {
            Class[] paramClass = { java.lang.Integer.TYPE, ownerName.getClass(), ownerMail.getClass() };
            Method m=c.getMethod("createKeys", paramClass );
	    Object[] param = {new Integer(size) , ownerName, ownerMail };
            kp = (KeyPair)m.invoke(null, param);
        } catch (Exception e) {
            System.err.println(e);
            throw new BadKeyClassException(name);
        }
        // kp is a new KeyPair as specified.
        return kp;
    }
    
    // CLASS VARIABLES
    
    /**
     * Random is used by descendants of this class to generate
     * random numbers. It should be a random number without a
     * period, which rules out all generators based on iterated
     * functions, such as java.util.Random.
     * <p>
     * If this field is null, an instance of RandomSpinner will
     * be created and used.
     *
     * @see IS.hi.logir.cryptonite.RandomSpinner */
    public static Random random=null;
    
    /**
     * We allow a chance of 0.5**primeCertainty chance that the primes we
     * generate are in fact composite numbers.
     * <p>
     * The default value of 32 will give a false prime less than once in every
     * 4,000,000,000 tries. However, finding one prime often takes several tries,
     * but the likelyhood of a false prime being returned is still negligeble. */
    public static int primeCertainty=32;
    
    /** The constant one. */
    protected static final BigInteger ONE=BigInteger.valueOf(1);
    
    /** The constant zero. */
    protected static final BigInteger ZERO=BigInteger.valueOf(0);

    // INSTANCE VARIABLES
    
    /** The key-owner's name. */
     private String ownerName;

    /** The key-owner's e-mail address. */
     private String ownerMail;
    
    /**
     * The key's MD5 fingerprint. It is never stored with the key,
     * but re-calculated at run-time. fingerprint may be null in
     * which case it needs to be re-calcualted. */
    private transient Fingerprint fingerprint;

    /**
     * The signatures vector may contain certificates for this key.
     * <p>
     * Never sign a key unless you have first-hand knowledge that it is valid!
     * <p>
     * A key which is not certified by a trusted party should not be used.
     * 
     * @see IS.hi.logir.cryptonite.Signature
     */
    private Vector signatures;

    // INSTANCE METHODS
    
    /**
     * Create a new key. It will be marked as belonging
     * to ownerName with e-mail address ownerMail. */
    public Key(String ownerName, String ownerMail){
        this.ownerName = ownerName;
        this.ownerMail = ownerMail;
	    
        signatures = new Vector();

	if (random == null)
            random = new RandomSpinner();
    }
    
    /**
     * Return the "size" of the key. This is a measure of how 
     * difficult it is to break and is heavily dependant on
     * the algorithm used. In many cases it will be the
     * actual number of bits needed to store the key. */
    public abstract int getSize();

    /** Return the name of the owner. */
    public final String getOwnerName(){
        return ownerName;
    }
    
    /** Return the e-mail address of the owner. */
    public final String getOwnerMail(){
        return ownerMail;
    }
    
    /**
     * Return the key's MD5 fingerprint.
     * 
     * @see IS.hi.logir.cryptonite.MD5State
     * @see IS.hi.logir.cryptonite.Fingerprintable
     */
    protected abstract Fingerprint calcFingerprint();
    
    /** Return the key's MD5 fingerprint. */
    public final Fingerprint getFingerprint(){
        if (fingerprint==null)
            fingerprint = calcFingerprint();
        return fingerprint;
    }

    /**
     * Return an enumeration of this keys certificates. These are of 
     * the Signature class.
     * 
     * @see IS.hi.logir.cryptonite.Signature */
     public Enumeration getSignatures(){
        return signatures.elements();
    }
    
    /** Return the name of the algorithm used by this key. */
    public abstract String algorithmName();

    /** Return a string representation of this key. */
    public String toString(){
        return algorithmName()+ " key ("+getFingerprint()+
	    ")\n    owned by "+getOwnerName()+" ("+getOwnerMail()+")";
    }
    
    /** Return a hash-code based on the MD5 fingerprint. */
    public int hashCode(){
        return getFingerprint().hashCode();
    }
    
    /**
     * Returns true if this and key are a matched pair
     * of public/private keys. */
    public abstract boolean matches(Key key);

    /**
     * Returns a number n, such that least space will be wasted if
     * a mutiple of n bytes is encrypted at once. */
    public int blockSize(){
	return 1;
    }
    
    /**
     * Encrypt source with this key and put the result in dest.
     * The source.buf and dest.buf arrays must be different. */
    public abstract void encrypt(Data source, Data dest);
    
    /**
     * Encrypt the first sourceLength bytes of source with
     * this key and return an array with the encrypted data.
     * This method ultmately calls encrypt(Data,Data), but causes
     * more memory allocation and garbage collection than using
     * that method directly. */
    public final byte[] encrypt(byte[] source, int sourceLength){
	Data s = new Data(source,sourceLength);
	Data d = new Data();
	encrypt(s,d);
	if (d.n>d.buf.length){
	    // There are extra bytes in the array.
	    byte [] buf = new byte[d.n];
	    System.arraycopy(d.buf,0, buf,0, d.n);
	    return buf;
	} else
	    return d.buf;
    }
    
    /** 
     * Encrypt source with this key and return an array with
     * the encrypted data.
     * This method ultmately calls encrypt(Data,Data), but causes
     * more memory allocation and garbage collection than using
     * that method directly. */
    public final byte[] encrypt(byte[] source){
        return encrypt(source, source.length);
    }
    
    /**
     * Decrypt the data from source with this key and put the plain data
     * in dest. The source.buf and dest.buf arrays must be different. */
    public abstract void decrypt(Data source, Data dest);

    /**
     * Decrypt the data from source with this key and return an array
     * with the plain bytes.
     * This method ultmately calls decrypt(Data,Data), but causes
     * more memory allocation and garbage collection than using
     * that method directly. */
    public byte[] decrypt(byte[] source){
	Data s = new Data(source);
	Data d = new Data();
	decrypt(s,d);
	if (d.buf.length > d.n){
            // There are extra bytes in the array.
	    byte [] buf = new byte[d.n];
	    System.arraycopy(d.buf,0, buf,0, d.n);
	    return buf;
	} else
	    return d.buf;
    }

}
