package IS.hi.logir.cryptonite;

import java.math.BigInteger;
import java.util.Random;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Date;
import java.io.Serializable;

/**
 * This class holds a collection of public keys the user has collected, along with
 * his private keys.
 * <p>
 * It contains methods to  find keys given the name and/or e-mail address of the owner,
 * and various key-validation and management routines.
 *
 * @see IS.hi.logir.cryptonite.Key
 * @see IS.hi.logir.cryptonite.KeyPair
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */

public final class KeyRing implements Serializable{

    // lastCheck is the date at which the fingerprints in the key-ring
    // were last checked. If lastCheck is null, then the class has
    // just been deserialized and all the MD5 signatures are suspect.
    private transient Date lastCheck=null;
    
    // pubKeys stores the public key's in the ring.
    // The hash-key is the key's fingerprint.
    private Hashtable pubKeys;
    
    // trustKeys stores the public key's in the ring which the owner of the
    // KeyRing has decided can be trusted to validate other keys. All the keys
    // in trustKeys are also stored in pubKeys. The hash-key is the key's
    // fingerprint.
    private Hashtable trustKeys;
    
    // priKeys stores the private keys belonging to to the KeyRing's owner.
    // The hash-key is the fingerprint of the corresponding public key.
    private Hashtable priKeys;
    
    /** Create a new empty KeyRing. */
    public KeyRing(){
        pubKeys   = new Hashtable();
        trustKeys = new Hashtable();
        priKeys   = new Hashtable();
    }

    /** Add a public Key to the KeyRing. */
    public void addPub(Key k){
        pubKeys.put(k.getFingerprint(), k);
    }
    
    /** Add a trusted Key to the KeyRing. */
    public void addTrusted(Key k){
        pubKeys.put(k.getFingerprint(), k);
        trustKeys.put(k.getFingerprint(), k);
    }

    /** Add a public/private Key-pair to the KeyRing. */
    public void addPair(Key pub, Key pri){
        pubKeys.put(pub.getFingerprint(), pub);
        priKeys.put(pub.getFingerprint(), pri);
    }
    
    /** Add a public/private Key-pair to the KeyRing. */
    public void addPair(KeyPair kp){
        addPair(kp.getPublic(), kp.getPrivate());
    }
    
    /** Add a trusted public/private Key-pair to the KeyRing. */
    public void addTrustedPair(Key pub, Key pri){
        pubKeys.put  (pub.getFingerprint(), pub);
        trustKeys.put(pub.getFingerprint(), pub);
        priKeys.put  (pub.getFingerprint(), pri);
    }
    
    /** Add a trusted public/private Key-pair to the KeyRing. */
    public void addTrustedPair(KeyPair kp){
        addTrustedPair(kp.getPublic(), kp.getPrivate());
    }
    
    /**
     * Return the public key with the specified fingerprint. If it does
     * not exist, null is returned instead. */
    public Key getPub(Fingerprint fingerprint) throws KeyRingCorruptException{
        testFingerprints();
        return (Key)pubKeys.get(fingerprint);
    }
    
    /**
     * Return the private key with the specified fingerprint. If it does
     * not exist, null is returned instead. */
    public Key getPri(Fingerprint fingerprint){
        Key k=(Key)priKeys.get(fingerprint);
        return k;
    }
    
    /**
     * Return the trusted key with the specified fingerprint. If it does
     * not exist, null is returned instead. */
    public Key getTrusted(Fingerprint fingerprint){
        Key k=(Key)priKeys.get(fingerprint);
        return k;
    }
    
    /** Returns true if a signature with the Key k is accepted as a key validation. */
    public boolean isTrusted(Key k){
        // return true if a key with the same fingerprint exists in the trusted ring.
        return (trustKeys.get(k.getFingerprint()) != null);
    }
    
    /**
     * Returns true if the Key k has a signature from a trusted party. This code is
     * dependant on the Fingerprint class which has not been implemented. */
    public boolean isValid(Key k) throws KeyRingCorruptException{
        Enumeration E=k.getSignatures();
        while(E.hasMoreElements()){
            // None of the elements we've looked at in S are a good signature.
            Signature sig = (Signature)E.nextElement();
            if (sig.check(k, this).isGood())
                return true;
        }
        return false;
    }

    // Check the MD5 fingerprints of all the keys stored in this key-ring.
    private void testFingerprints() throws KeyRingCorruptException{
        Enumeration E=pubKeys.keys();
        while (E.hasMoreElements()){
            Fingerprint f=(Fingerprint)E.nextElement();
            Key k=(Key)pubKeys.get(f);
            if (!f.equals(k.getFingerprint()))
                throw new KeyRingCorruptException("Key-ring signatures have been tampered with! (public)");
        }
        // All the fingerprints in pubKeys have been verified.

        E=trustKeys.keys();
        while (E.hasMoreElements()){
            Fingerprint f=(Fingerprint)E.nextElement();
            Key k=(Key)trustKeys.get(f);
            if (!f.equals(k.getFingerprint()))
                throw new KeyRingCorruptException("Key-ring signatures have been tampered with! (trusted)");
        }
        // All the fingerprints in trustKeys have been verified.

        E=priKeys.keys();
        while (E.hasMoreElements()){
            Fingerprint f=(Fingerprint)E.nextElement();
            Key pub=(Key)pubKeys.get(f);
            Key pri=(Key)priKeys.get(f);
            if (pub==null || pri==null || !pub.matches(pri))
                // one of the keys is missing or theyare not a public/private pair
                throw new KeyRingCorruptException("Key-ring signatures have been tampered with! (private)");
        }
        // All the fingerprints in priKey have been verified.
        
        lastCheck=new Date();
    }
    
    /** Remove a key from the KeyRing. */
    public void remove(Key k){
        pubKeys.remove(k.getFingerprint());
        trustKeys.remove(k.getFingerprint());
        priKeys.remove(k.getFingerprint());
    }
    
    // See findPub() and findPri()
    private static Key find(String name, String mail, Hashtable keys){
        Enumeration e = keys.elements();
        
        while (e.hasMoreElements()){
            Key key = (Key)e.nextElement();
            if (name==null) {
                if (mail==null)   // no criterion
                    return key;
                else {            // only mail criterion
                    if (key.getOwnerMail().indexOf(mail)>=0)
                        return key;
                }
            } else {
                if (mail==null){  // only name
                    if (key.getOwnerName().indexOf(name)>=0)
                        return key;
                } else {          // both name and mail criteria
                    if ( (key.getOwnerName().indexOf(name)>=0) && (key.getOwnerMail().indexOf(mail)>=0) )
                        return key;
                }
            }
        }
        return null;
    }
    
    /**
     * Find a public key partially matching a name/e-mail pair. If name is not null,
     * then it will be a sub-string of the name of the owner of the key. If mail is
     * not null, then it will be a sub-string of the e-mail address of the owner of
     * the key.
     * <p>
     * If no such key exists, null is returned.
     * <p>
     * If more than one key mathces the criteria, the first one found is returned.*/
    public Key findPub(String name, String mail){
        return find(name, mail, pubKeys);
    }
    
    /**
     * Find a private key partially matching a name/e-mail pair. If name is not null,
     * then it will be a sub-string of the name of the owner of the key. If mail is
     * not null, then it will be a sub-string of the e-mail address of the owner of
     * the key.
     * <p>
     * If no such key exists, null is returned.
     * <p>
     * If more than one key mathces the criteria, the first one found is returned.*/
    public Key findPri(String name, String mail){
        return find(name, mail, priKeys);
    }
    
    // See findListPub() and findListPri()
    private static Key[] findList(String name, String mail, Hashtable keys){
        Enumeration e = keys.elements();
        Key[] list = new Key[keys.size()];
        int listPos=0;
        
        while (e.hasMoreElements()){
            Key key = (Key)e.nextElement();
            if (name==null) {
                if (mail==null)   // no criterion
                    list[listPos++] = key;
                else {            // only mail criterion
                    if (key.getOwnerMail().indexOf(mail)>=0)
                        list[listPos++] = key;
                }
            } else {
                if (mail==null){  // only name
                    if (key.getOwnerName().indexOf(name)>=0)
                        list[listPos++] = key;
                } else {          // both name and mail criteria
                    if ( (key.getOwnerName().indexOf(name)>=0) && (key.getOwnerMail().indexOf(mail)>0) )
                        list[listPos++] = key;
                }
            }
        }
        
        if (listPos<list.length){
            Key[] tmp=list;
            list = new Key[listPos];
            System.arraycopy(tmp,0, list,0, listPos);
        }
        return list;
    }
    
    /**
     * Find all public keys partially matching a name/e-mail pair. If name is not null,
     * then it will be a sub-string of the name of the owner of the keys. If mail is not
     * null, then it will be a sub-string of the e-mail address of the owner of the keys. */
    public Key[] findListPub(String name, String mail){
        return findList(name, mail, pubKeys);
    }
    
    /**
     * Find all private keys partially matching a name/e-mail pair. If name is not null,
     * then it will be a sub-string of the name of the owner of the keys. If mail is not
     * null, then it will be a sub-string of the e-mail address of the owner of the keys. */
    public Key[] findListPri(String name, String mail){
        return findList(name, mail, priKeys);
    }
}