package IS.hi.logir.cryptonite;

import java.util.Random;

/**
 * This class uses the scheduler to generate random numbers. It counts the number of times
 * a loop is repeated before a thread has slept for a specified number of milliseconds.
 * The two lowest order bits of this number are then used.
 * <p>
 * The <a href="http://www.cs.berkeley.edu/~daw/java-spinner2">helper class</a>
 * which does the actual number generation is by
 * Henry Strickland (<a href="strix@versant.com">strix@versant.com</a>) and
 * Greg Noel (<a href="greg@qualcomm.com">greg@qualcomm.com</a>). It is based on
 * <a href="ftp://ftp.research.att.com/dist/mab/librand.shar">similar C code</a>
 * by Matt Blaze, Jack Lacy, and Don Mitchell.
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class RandomSpinner extends Random {

    public RandomSpinner(){
    }
    
    /** Generates the next random number. */
    protected synchronized int next(int bits){
        return (int)
            ((Spinner.spin(10) & 0x3) << 30
            |(Spinner.spin(10) & 0x3) << 28
            |(Spinner.spin(10) & 0x3) << 26
            |(Spinner.spin(10) & 0x3) << 24
            |(Spinner.spin(10) & 0x3) << 22
            |(Spinner.spin(10) & 0x3) << 20
            |(Spinner.spin(10) & 0x3) << 18
            |(Spinner.spin(10) & 0x3) << 16
            |(Spinner.spin(10) & 0x3) << 14
            |(Spinner.spin(10) & 0x3) << 12
            |(Spinner.spin(10) & 0x3) << 10
            |(Spinner.spin(10) & 0x3) <<  8
            |(Spinner.spin(10) & 0x3) <<  6
            |(Spinner.spin(10) & 0x3) <<  4
            |(Spinner.spin(10) & 0x3) <<  2
            |(Spinner.spin(10) & 0x3));
    }

    private static final int NUM_ITER=256;  // should be a multiple of 16

    /**
     * Test randomness. This is done by applying a chi-square test to the 2 lowest bits
     * of the values returned by the Spinner helper class and by applying a chi-square
     * test to the lengths of the gaps between identical values in the random stream.
     */
    public static boolean isRandom(){
        
        int[] count=new int[4];
        int[] last=new int[4];
        int[] gapCount=new int[10];

        Random rnd = new Random();

        for(int i=0; i<last.length; i++)
            last[i]=-1;

        for(int i=0; i<NUM_ITER; i++){
            // The value j has appeared count[j] times.
            // The value j last appeared at iteration last[j]
            // gaps of length j have appeared gapCount[j] times.
            int n = (int)(Spinner.spin(10) & 0x3);
            count[n]++;
            gapCount[Math.min(i-last[n]-1,9)]++;
            last[n]=i;
        }

        double x;
        double V=0;
        for (int j=0; j<4; j++){
            x=count[j]-(NUM_ITER/4);
            V += (x*x)/(NUM_ITER/4);
        }
        
        double np = NUM_ITER*0.25;
        double W = 0;
        for (int i=0; i<9; i++){
            x=gapCount[i]-np;
            W += (x*x)/np;
            np *= 0.75;
        }
        np /= 0.25;
        x=gapCount[9]-np;
        W += (x*x)/np;
        
        System.out.println("V = "+V);
        System.out.println("W = "+W);
        return (0.3518<V) && (V<7.815) && (2.088<W) && (W<21.67);
    }
    
}