package IS.hi.logir.cryptonite;

import java.math.BigInteger;
import java.io.Serializable;

/**
 * The RSA algorithm is probably the best known and most widely used
 * public key algorithm. Breaking one RSA key is exactly as difficult
 * as factoring the large integer that comprises the key, and there is
 * no known way to do this in a reasonable time. Therefore RSA should
 * be about as secure as anything if you keep your keys long. 1024 bits
 * should be more than enough in most cases, but the truly paranoid may
 * want to use up to 4096 bit keys.
 * <p>
 * Each RSA key is a pair (r,n) of integers and matches another key (s,n).
 * If P is a block of plain data represented as an integer smaller than n,
 * then it can be encrypted with the transformation:
 * <blockquote>
 *   <tt>E = (P^r) mod n</tt>
 * </blockquote>
 * which has the inverse transformation:
 * <blockquote>
 *   <tt>P = (E^s) mod n</tt>
 * </blockquote>
 * <p>
 * The key owner will keep one key secret and publish the other as widely
 * as possible. This allows anyone who gets hold of the public key to
 * encrypt data which can only be decrypted with the corresponding private
 * key.
 * <p>
 * Data that is encrypted with a private key can similarly only be
 * decrypted with the corresponding public key. This is useful for digital
 * signatures.
 * <p>
 * When P is created from an array of bytes, it will correspond to as many
 * bytes of plain data as the bytes needed to store n, less one.
 * <p>
 * When the encrypted integer E is converted to a byte array, it will
 * contain as many bytes as are needed to store n. This means
 * that the encrypted data will be slightly larger, by a ratio of
 * (m+1)/m, with m=trunc(log2(n)/8).
 * <p>
 * Each chunk of data encrypted with RsaKey has as many bytes as the key modulo. However,
 * the encrypted data it contains corresponds to plain data with one less byte.
 *
 * @see IS.hi.logir.cryptonite.Signature
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public final class RsaKey extends Key implements Serializable, Fingerprintable{

    /** R is the exponent used in all public keys. */    
    private static final BigInteger R=BigInteger.valueOf(23);
    
    /** The RSA key exponent */
    private BigInteger r;
    
    /** The RSA key modulo */
    private BigInteger n;

    /**
     * Create a new RSA key (r,n). It will be marked as belonging
     * to ownerName with e-mail address ownerMail. */    
    public RsaKey(BigInteger r, BigInteger n, String ownerName, String ownerMail) {
        super(ownerName,ownerMail);
        this.r = r;
        this.n = n;
    }
    
    /**
     * Create a pair of public/private keys. The key modulo will be at most
     * bitLength bits in length and not much shorter. It will be marked as
     * belonging to ownerName with e-mail address ownerMail. */    
    public static KeyPair createKeys(int bitLength, String ownerName, String ownerMail){
        if (random==null)
            random=new RandomSpinner();
	if (bitLength<16)
	    bitLength=16;

        BigInteger s=null;
        BigInteger n=null;
        do {
            try{
                BigInteger p=new BigInteger(bitLength/2, primeCertainty, random);
                BigInteger q=new BigInteger(bitLength/2, primeCertainty, random);
                BigInteger m=p.subtract(ONE).multiply(q.subtract(ONE));
                s=R.modInverse(m);
                n=p.multiply(q);
            } catch (ArithmeticException e) {
                // Key generation failed... just try again.
                n=null;
            }
        } while(n==null);

        Key pub=new RsaKey(R,n, ownerName, ownerMail);
        Key pri=new RsaKey(s,n, ownerName, ownerMail);
        return new KeyPair(pub,pri);
    }

    /**
     * Create a KeyPair object holding objects for the public RSA key (r,n)
     * and the private RSA key (s,n). They will be marked as belonging to
     * ownerName with e-mail address ownerMail. */    
    public static KeyPair createKeys(BigInteger r, BigInteger s, BigInteger n, String ownerName, String ownerMail) throws KeyException {
	Key pub = new RsaKey(r,n, ownerName, ownerMail);
	Key pri = new RsaKey(s,n, ownerName, ownerMail);
	if (!pub.matches(pri))
	    throw new KeyException("The (r,s,n) triplet does not specify a matching pair of keys.");
	return new KeyPair(pub,pri);
    }

    /** Return the size of the key modulo in bits. */
    public int getSize(){
        return n.bitLength();
    }

    /**
     * Return the key's MD5 fingerprint.
     * 
     * @see IS.hi.logir.cryptonite.MD5State
     * @see IS.hi.logir.cryptonite.Fingerprintable
     */
    protected final Fingerprint calcFingerprint(){
        byte[] br=r.toByteArray();
        byte[] bn=n.toByteArray();
        byte[] bN=getOwnerName().getBytes();
        byte[] bM=getOwnerMail().getBytes();
        byte[] b = new byte[br.length+bn.length+bN.length+bM.length];
        System.arraycopy(br,0, b,0                            , br.length);
        System.arraycopy(bn,0, b,br.length                    , bn.length);
        System.arraycopy(bN,0, b,br.length+bn.length          , bN.length);
        System.arraycopy(bM,0, b,br.length+bn.length+bN.length, bM.length);
        return MD5State.calcFingerprint(b);
    }

    /** Return the name of the algorithm used by this key. */
    public String algorithmName(){
        return ("RSA");
    }

    /** Return a string representation of this key. */
    public String toString(){
	return getSize()+" bit RSA key ("+getFingerprint()+
	    ")\n    from "+getOwnerName()+" ("+getOwnerMail()+")";
    }

    /** Return true if the two keys are eqivalent. */
    public final boolean equals(Object o){
        if (o.getClass() != this.getClass())
            return false;
        RsaKey key=(RsaKey)o;
        return ( (key.r.equals(r)) &&
                 (key.n.equals(n)) &&
                 (key.getOwnerName().equals(getOwnerName())) &&
                 (key.getOwnerMail().equals(getOwnerMail())) );
    }

    /** 
     * Returns true if this and key are a matched pair of
     * public/private keys. */
    public final boolean matches(Key key){
        if (key.getClass() != this.getClass())
            return false;
        RsaKey k=(RsaKey)key;
        if (!n.equals(k.n))
            return false;
        int bl = n.bitLength();
        for (int i=0; i<10; i++){
            BigInteger a=new BigInteger(bl-1, random);
            BigInteger b=a.modPow(k.r,n).modPow(r,n);
            if (!a.equals(b))
                return false;
        }
        return true;
    }

    /**
     * Returns a number n, such that least space will be wasted if
     * a mutiple of n bytes is encrypted at once. */
    public int blockSize(){
	return (n.bitLength()-1)/8;
    }
    
    /**
     * Encrypt source with this key and put the result in dest.
     * The source.buf and dest.buf arrays must be different. */
    public final void encrypt(Data source, Data dest){
	// size of each chunk to encrypt
       	int chunkSize=(n.bitLength()-1)/8;

	// number of full-size chunks
        int numChunks=source.n/chunkSize;  

	// small chunk at the end of input?
        boolean smallChunk = source.n % chunkSize != 0; 

	// minimum size of destination buffer
	int destLength = 4+(chunkSize+1)*(numChunks+(smallChunk?1:0));

        byte[] chunk = new byte[chunkSize];
	dest.enlarge(destLength);
	dest.n = destLength;
	
        dest.buf[0]=(byte)(source.n >>> 24);
        dest.buf[1]=(byte)(source.n >>> 16);
        dest.buf[2]=(byte)(source.n >>>  8);
        dest.buf[3]=(byte)source.n;

        for(int i=0; i<numChunks; i++){
	    // The first i chunks have been encrypted and written to dest
            System.arraycopy(source.buf,i*chunkSize, chunk,0, chunkSize);
            byte[] chunk2 = new BigInteger(1,chunk).modPow(r,n).toByteArray();
            if(chunk2.length>=chunkSize+1) 
		// The encrypted array is at least chunkSize+1 bytes in length
                System.arraycopy(chunk2,chunk2.length-(chunkSize+1), dest.buf, 4+i*(chunkSize+1), chunkSize+1);
            else 
		// The encrypted array is a bit on the short side
                System.arraycopy(chunk2,0, dest.buf, 4+(i+1)*(chunkSize+1)-chunk2.length, chunk2.length);
        }
        if(smallChunk){
	    // There is a smaller chunk at the end which we need to encrypt.
	    int lastLength = source.n%chunkSize;
	    chunk = new byte[lastLength];
            System.arraycopy(source.buf,numChunks*chunkSize, chunk,0, lastLength);
            byte[] chunk2 = new BigInteger(1,chunk).modPow(r,n).toByteArray();
            if(chunk2.length>=chunkSize+1)
		// The encrypted array is at least chunkSize+1 bytes in length
                System.arraycopy(chunk2,chunk2.length-(chunkSize+1), dest.buf, 4+numChunks*(chunkSize+1), chunkSize+1);
            else
		// The encrypted array is a bit on the short side
                System.arraycopy(chunk2,0, dest.buf, 4+(numChunks+1)*(chunkSize+1)-chunk2.length, chunk2.length);
        }
    }
    
    /**
     * Decrypt the data from source with this key and put the plain data
     * in dest. The source.buf and dest.buf arrays must be different. */
    public void decrypt(Data source, Data dest){
        int chunkSize=(n.bitLength()-1)/8+1;
	int headerSize=4;
        int destLength
            = (source.buf[0]<0 ? 0x100+source.buf[0] : source.buf[0]) << 24
            | (source.buf[1]<0 ? 0x100+source.buf[1] : source.buf[1]) << 16
            | (source.buf[2]<0 ? 0x100+source.buf[2] : source.buf[2]) <<  8
            | (source.buf[3]<0 ? 0x100+source.buf[3] : source.buf[3]);
	
	dest.enlarge(destLength);
        dest.n=destLength;
	
        int numChunks=(destLength)/(chunkSize-1);    
	// number of full-size chunks
        boolean smallChunk = destLength % (chunkSize-1) != 0;  
	// small chunk at the end of input
	
        byte[] chunk = new byte[chunkSize];
        
        for(int i=0; i<numChunks; i++){     
	    // The first i chunks have been decrypted and written to dest
            System.arraycopy(source.buf,headerSize+i*chunkSize, chunk,0, chunkSize);
            byte[] chunk2 = new BigInteger(1,chunk).modPow(r,n).toByteArray();
            if(chunk2.length >= chunkSize-1)
		// The decrypted array is at least chunkSize-1 bytes in length
                System.arraycopy(chunk2,chunk2.length-(chunkSize-1), dest.buf,i*(chunkSize-1), chunkSize-1);
            else
		// The decrypted array is a bit on the short side
                System.arraycopy(chunk2,0, dest.buf,(i+1)*(chunkSize-1)-chunk2.length, chunk2.length);
        }
        if(smallChunk){
	    // The last chunk contains some padding
            int lastLength=destLength % (chunkSize-1);
            System.arraycopy(source.buf,headerSize+numChunks*(chunkSize), chunk,0, chunkSize);
            byte[] chunk2 = new BigInteger(1,chunk).modPow(r,n).toByteArray();
            if(chunk2.length >= lastLength) {
		// The decrypted array is at least lastLength bytes in length
		System.arraycopy(chunk2,chunk2.length-lastLength, dest.buf,numChunks*(chunkSize-1), lastLength);
	    } else {
		// The decrypted array is a bit on the short side
                System.arraycopy(chunk2,0, dest.buf,dest.n-chunk2.length, chunk2.length);
	    }
        }
    }
    
}
