package IS.hi.logir.cryptonite;

import java.io.Serializable;

/**
 * This class stores an rsa signature. A signature of a chunk of data is an
 * MD5 fingerprint of that data encrypted with a private key. This signature
 * can be verified with the corresponding public key. Faking a signature is
 * exactly as difficult as decrypting a message that uses the same key pair.
 * <p>
 * In particular, if Alice signs a key which claims to belong to Bob, she's
 * confirming that it does indeed belong to Bob. Thus if Carl recieves it
 * and trusts Alice, he can safely use Bob's key.
 *
 * @see IS.hi.logir.cryptonite.Fingerprint
 * @see IS.hi.logir.cryptonite.RsaKey
 *
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class Signature implements Serializable{

    private byte[] s;             // encrypted fingerprint.
    private Fingerprint keyPrint; // Fingerprint of the key used to create the signature.

    /** Create a new signature for obj, using the specified private key. */
    public Signature(Fingerprintable obj, Key key){
        Fingerprint f = obj.getFingerprint();
        s = key.encrypt(f.getBytes());
    }

    /** Create a new signature for the byte array, using the specified private key. */
    public Signature(byte[] buf, Key key){
        Fingerprint f = MD5State.calcFingerprint(buf);
        s = key.encrypt(f.getBytes());
    }

    /** Return the fingerprint of the key used to create this signature. */
    public Fingerprint getKeyFingerprint(){
        return keyPrint;
    }

    /**
     * Check if this is a signature for obj created with the private
     * counterpart to the specified trusted public key. */
    public SignatureMatch check(Fingerprintable obj, Key key){
        if (key==null)
            return new SignatureMatch(false, false, false);
        
        byte[] s1 = obj.getFingerprint().getBytes();
        byte[] s2 = key.encrypt(s);
        // s1 should now match s2 if this signature is valid.
        
        if (s1.length!=s2.length)
            return new SignatureMatch(true, true, false);
        for (int i=s1.length-1; i>=0; i--)
            if (s1[i]!=s2[i])
                return new SignatureMatch(true, true, false);
        return new SignatureMatch(true, true, true);
    }
    
    /**
     * Check if this is a signature for buf created with the private
     * counterpart to the specified trusted public key. */
    public SignatureMatch check(byte[] buf, Key key){
        byte[] s1 = MD5State.calcFingerprint(buf).getBytes();
        byte[] s2 = key.encrypt(s);
        // s1 should now match s2 if this signature is valid.
        
        if (s1.length!=s2.length)
            return new SignatureMatch(true, true, false);
        for (int i=s1.length-1; i>=0; i--)
            if (s1[i]!=s2[i])
                return new SignatureMatch(true, true, false);
        return new SignatureMatch(true, true, true);
    }
    
    /**
     * Check if this is a signature for obj created with the private
     * counterpart to a trusted public key from the specified key-ring. */
    public SignatureMatch check(Fingerprintable obj, KeyRing keyRing) throws KeyRingCorruptException{
        Key key=keyRing.getTrusted(keyPrint);
        if (key==null) {
            key=keyRing.getPub(keyPrint);
            SignatureMatch sm=check(obj, key);
            sm.keyTrusted=false;
            return sm;
        } else
            return check(obj, key);
    }
    
    /**
     * Check if this is a signature for buf created with the private
     * counterpart to a trusted public key from the specified key-ring. */
    public SignatureMatch check(byte[] buf, KeyRing keyRing) throws KeyRingCorruptException{
        Key key=keyRing.getTrusted(keyPrint);
        if (key==null) {
            key=keyRing.getPub(keyPrint);
            SignatureMatch sm=check(buf, key);
            sm.keyTrusted=false;
            return sm;
        } else
            return check(buf, key);
    }
    
}
