package IS.hi.logir.cryptonite;

import java.io.Serializable;

/************************************************
 * The Tiny Encryption Algorithm (TEA) by David Wheeler and Roger Needham
 * of the <a href="http://vader.brad.ac.uk/tea.html">Cambridge Computer
 * Laboratory</a>.
 * <p>
 * TEA is a Feistel cipher with XOR and and addition as the non-linear
 * mixing functions. 
 * <p>
 * The number of iterations is variable. 32 is ample, 16 is sufficient,
 * as few as eight may be OK. The algorithm achieves good dispersion
 * after six iterations.
 * <p>
 * @author <a href="http://www.hi.is/~logir/">Logi Ragnarsson</a> (<a href="mailto:logir@hi.is">logir@hi.is</a>)
 */
public class TeaKey extends Key implements Serializable, Fingerprintable{
    /*
     * Notes: (in more detail)
     * 
     * The Tiny Encryption Algorithm (TEA) by
     * David Wheeler and Roger Needham of the
     * Cambridge Computer Laboratory.
     * 
     * TEA is a Feistel cipher with XOR and
     * and addition as the non-linear mixing
     * functions. 
     * 
     * Takes 64 bits of data in v[0] and v[1].
     * Takes 128 bits of key in k[0] - k[3].
     *
     * TEA can be operated in any of the modes
     * of DES. Cipher Block Chaining is, for example,
     * simple to implement.
     * 
     * n is the number of iterations. 32 is ample,
     * 16 is sufficient, as few as eight may be OK.
     * The algorithm achieves good dispersion after
     * six iterations. The iteration count can be
     * made variable if required.
     *
     * Note this is optimised for 32-bit CPUs with
     * fast shift capabilities. It can very easily
     * be ported to assembly language on most CPUs.
     * 
     * delta is chosen to be the real part of the
     * golden ratio, Sqrt(5/4) + 1/2 ~ 1.618034
     * multiplied by 2^32. 
     ************************************************/
    
    final private static long delta = 0x9E3779B9L;
    final private static long intMask = 0x00ffffffffL;
    final private static int charBits = 16;
    final private static int byteBits = 8;
    final private static int charMask = 0xffff;
    final private static int byteMask = 0xff;
    
    private long key[];
    private int iter;
    private int logIter;
    private long sumEnd;
    
    /**
     * Create a new random Tea key with keySize iterations. It will
     * be marked as belonging to ownerName with e-mail address
     * ownerMail. */
     public TeaKey(int keySize, String ownerName, String ownerMail){
	 super(ownerName,ownerMail);
	 setIter(keySize);

 	 key = new long[4];
	 key[0] = ((long)random.nextInt()) & intMask;
	 key[1] = ((long)random.nextInt()) & intMask;
	 key[2] = ((long)random.nextInt()) & intMask;
	 key[3] = ((long)random.nextInt()) & intMask;
     }
    
    /**
     * Create an object for the specified Tea key. It will
     * be marked as belonging to ownerName with e-mail address
     * ownerMail. */
     public TeaKey(int keySize, long k1, long k2, long k3, long k4, String ownerName, String ownerMail){
	 super(ownerName,ownerMail);
	 setIter(keySize);

 	 key = new long[4];
	 key[0] = k1 & intMask;
	 key[1] = k2 & intMask;
	 key[2] = k3 & intMask;
	 key[3] = k4 & intMask;
     }
    
    // sets iterations to the next power of two, that is <= n
    private void setIter(int n){
	if (n<8)
	    n=8;
	logIter = log2(n);
	iter = 1 << logIter;
	sumEnd = (delta << logIter) & intMask; //delta*iter
    }

    /**
     * Create a TEA random key-pair. Both keys will be identical, since TEA
     * is a symmetric cipher. The key will perform keySize iterations during
     * encryption. It will be marked as belonging to ownerName with e-mail
     * address ownerMail. */
    public static KeyPair createKeys(int keySize, String ownerName, String ownerMail){
	TeaKey t = new TeaKey(keySize, ownerName, ownerMail);
	return new KeyPair(t,t);
    }
        
    private int log2(int n) {
	if (n == 0) 
	    throw new ArithmeticException("log2 of zero");
	
	int i = -1;
	while (n != 0){
	    ++i;
	    n = n >>> 1;
	}
	return i;
    }

    /**
     * Returns the number of iterations performed when encrypting
     * with this key. */
    public int getSize(){
	return iter;
    }

    /**
     * Return the key's MD5 fingerprint. It is called by the final
     * getFingerprint method in the Key superclass.
     *
     * @see IS.hi.logir.cryptonite.MD5State
     * @see IS.hi.logir.cryptonite.Fingerprintable
     */    
    protected final Fingerprint calcFingerprint(){
	byte[] bN=getOwnerName().getBytes();
        byte[] bM=getOwnerMail().getBytes();
        byte[] b = new byte[16+bN.length+bM.length];
        
        b[0] = (byte)((key[0]>>>(0*byteBits)) & byteMask); 
        b[1] = (byte)((key[0]>>>(1*byteBits)) & byteMask);
        b[2] = (byte)((key[0]>>>(2*byteBits)) & byteMask); 
        b[3] = (byte)((key[0]>>>(3*byteBits)) & byteMask);
        
        b[4] = (byte)((key[1]>>>(0*byteBits)) & byteMask); 
        b[5] = (byte)((key[1]>>>(1*byteBits)) & byteMask);
        b[6] = (byte)((key[1]>>>(2*byteBits)) & byteMask); 
        b[7] = (byte)((key[1]>>>(3*byteBits)) & byteMask);
        
        b[8] = (byte)((key[2]>>>(0*byteBits)) & byteMask); 
        b[9] = (byte)((key[2]>>>(1*byteBits)) & byteMask);
        b[10]= (byte)((key[2]>>>(2*byteBits)) & byteMask); 
        b[11]= (byte)((key[2]>>>(3*byteBits)) & byteMask);
        
        b[12]= (byte)((key[3]>>>(0*byteBits)) & byteMask); 
        b[13]= (byte)((key[3]>>>(1*byteBits)) & byteMask);
        b[14]= (byte)((key[3]>>>(2*byteBits)) & byteMask); 
        b[15]= (byte)((key[3]>>>(3*byteBits)) & byteMask);
        
        System.arraycopy(bN,0, b,16          , bN.length);
        System.arraycopy(bM,0, b,16+bN.length, bM.length);
        return MD5State.calcFingerprint(b);
    }
    
    /** Return the name of the algorithm used by this key. */
    public String algorithmName(){
	return "TEA"; 
    }
    
    /** Return a string representation of this key. */
    public String toString(){
        return "TEA key ("
            + Long.toHexString(key[0]) + Long.toHexString(key[1])
            + Long.toHexString(key[2]) + Long.toHexString(key[3])
            + ")\n    from "+getOwnerName()+" ("+getOwnerMail()+")";
    }


    /** Return true iff o is equal to this Key. */
    public final boolean equals(Object o){
        if (o.getClass() != this.getClass())
	    return false;
	
        TeaKey t = (TeaKey)o;
        if (t.key.length != 4)
	    return false;
        for (int i=0; i<4; i++){
            if (t.key[i] != key[i])
		return false;
        }
        return true;
    }

    /**
     * Returns true if this and key are a matched pair
     * of public/private keys. For symetric ciphers such
     * as TEA, this is true iff k is equal to this key. */
    public final boolean matches(Key k){
	return equals(k); 
    }
    
    /** Encrypt v[offset] and v[offset+1] in place */
    public void encryptElement(int v[], int offset){
	long y=((long)v[offset]  ) & intMask;
	long z=((long)v[offset+1]) & intMask;
	long sum=0;
	long a=key[0],	b=key[1];
	long c=key[2],	d=key[3];
	
	int n=iter;
	while ( n-- > 0){
	    sum = (sum+delta) & intMask;
	    y = (y + ( ((z<<4)+a) ^ (z+sum) ^ ((z>>>5)+b) ))
		& intMask;
	    z = (z + ( ((y<<4)+c) ^ (y+sum) ^ ((y>>>5)+d) ))
		& intMask;
	}
	v[offset]   = (int)y;
	v[offset+1] = (int)z;
    }
    
    /** Encrypt the array v in place */
    public void encryptArray(int [] v){
	int j = v.length-1;
	for (int i=0; i<j; i++)
	    encryptElement(v,i);
    }
    
    /** Decrypt v[offset] and v[offset+1] in place */
    public void decryptElement(int v[], int offset){
	long y=((long)v[offset]  ) & intMask;
	long z=((long)v[offset+1]) & intMask;
	long sum=sumEnd;
	int n=iter;
	long a=key[0],	b=key[1];
	long c=key[2],	d=key[3];
	
	while (n-- > 0){
	    z = (z - ( ((y<<4)+c) ^ (y+sum) ^ ((y>>>5)+d) ))
		& intMask;
	    y = (y - ( ((z<<4)+a) ^ (z+sum) ^ ((z>>>5)+b) ))
		& intMask;
	    sum = (sum - delta) & intMask;
	}
	v[offset]  =(int)y;
	v[offset+1]=(int)z;
    }
    
    /** Decrypt the array v in place. */
    public void decryptArray(int [] v){
	for (int i=v.length-2; i>=0; --i) decryptElement(v,i);
    }
	
    /** Put the bits from i in b[bOffset..bOffset+3]. */
    private void intToByte(int i, byte[]b, int bOffset){
	b[bOffset+0] = (byte)((i>>>(3*byteBits)) & byteMask);
	b[bOffset+1] = (byte)((i>>>(2*byteBits)) & byteMask);
	b[bOffset+2] = (byte)((i>>>(1*byteBits)) & byteMask);
	b[bOffset+3] = (byte)((i>>>(0*byteBits)) & byteMask);
    }
    
    /** Put the bits from b[bOffset..bOffset+3] into i[iOffset] */
    private void byteToInt(byte[]b, int bOffset, int[] i, int iOffset){
	int j=0;
	i[iOffset] = 0;
	if (bOffset >= b.length) return;
	j = ((int)b[bOffset]) & byteMask; // make sure sign doesn't extend
	i[iOffset] = i[iOffset] | (j << 3*byteBits);
         ++bOffset;
     
         if (bOffset >= b.length) return;
         j = ((int)b[bOffset]) & byteMask; // make sure sign doesn't extend
         i[iOffset] = i[iOffset] | (j << 2*byteBits);
         ++bOffset;
         
         if (bOffset >= b.length) return;
         j = ((int)b[bOffset]) & byteMask; // make sure sign doesn't extend
         i[iOffset] = i[iOffset] | (j << 1*byteBits);
         ++bOffset;
      
         if (bOffset >= b.length) return;
         j = ((int)b[bOffset]) & byteMask; // make sure sign doesn't extend
         i[iOffset] = i[iOffset] | (j << 0*byteBits);
         ++bOffset;     
     }

     /**
      * Returns a number n, such that least space will be wasted if
      * a mutiple of n bytes is encrypted at once. */
     public int blockSize(){
	 return 4;
     }

    /**
     * Encrypt source with this key and put the result in dest.
     * The source.buf and dest.buf arrays must be different. */
     public final void encrypt(Data source, Data dest){
	 int m = source.n/4;
	 int[] d;
	 if (source.n % 4 ==0)
	     d = new int[m+1];
	 else
	     d = new int[m+2];
	 int j = 0;
	 
	 d[j++] = source.n;
	 // d[0] is the length of the plain-text
	 for (int i=0; i < source.n; i+=4, j++)
	     byteToInt(source.buf, i, d, j);
	 // d[1..d.length-1] is the plain-text from source, possibly with
	 // appended 0's if the length of source was not a multiple of
	 // 4 bytes.
	 
	 encryptArray(d);
         dest.enlarge(4*d.length);
         dest.n=4*d.length;
         intToByte(source.n,dest.buf,0);  // arf essa lnun nokku?
         for (int i=0; i<d.length; i++)
             intToByte(d[i],dest.buf,4*i);
      }

    /**
     * Decrypt the data from source with this key and put the plain data
     * in dest. */
    public void decrypt(Data source, Data dest){
	if (source.n % 4 != 0)
	    throw new IllegalArgumentException(
		"Input byte array must have a length that is a multiple "+
	        "of 4. Array parameter has length "+source.n);
	int inputLen = source.n/4;
	if (inputLen < 2)
	    throw new StringIndexOutOfBoundsException("Input length "+source.n+" too small!");
	int[] v = new int[inputLen];
	for (int i=0; i<inputLen; i++)
	    byteToInt(source.buf,4*i,v,i);
	
	decryptArray(v);
	int l = v[0]; // size saved when we encrypted the array
	
	if ( (l <= source.n-8) || (l > source.n-4) )
	    throw new IllegalArgumentException
	    ("Encrypted length should be strictly greater than "
	     +(source.n-8)+" and less than or equal to "
	     +(source.n-4)+". The decrypted value is "+l);
	
	dest.enlarge( (l%4 ==0) ? (l) : (l/4+1)*4 );
	dest.n=l;
	for (int i=1; i<inputLen; i++)
	    intToByte(v[i],dest.buf,4*(i-1));
    }      
      
}

